from pyramid import tweens

__all__ = (
    '__version__',
    '__name__',  # pylint: disable=undefined-all-variable
)

__version__ = '0.1.0'


def includeme(config):
    """Includes all twees that are provided by `pyramid_secure_response`.

    >>> from pyramid.config import Configurator
    >>> config = Configurator(settings={})

    >>> config.include('pyramid_secure_response')

    Otherwise, use just `add_tween()` as you need.

    >>> config.add_tween('pyramid_secure_response.ssl_redirect.tween')
    >>> config.add_tween('pyramid_secure_response.hsts_support.tween')
    """
    tween_name = (lambda name: '{:s}.{:s}.tween'.format(__name__, name))

    config.add_tween(tween_name('ssl_redirect'),
                     over=tweens.MAIN)

    config.add_tween(tween_name('hsts_support'),
                     over=tweens.MAIN,
                     under=tween_name('ssl_redirect'))



from pyramid.httpexceptions import HTTPMovedPermanently

from pyramid_secure_response.util import (
    logger,
    apply_path_filter,
    get_config,
    build_criteria,
)


def tween(handler, registry):
    """Redirects insecure HTTP request as configured.

    This tween does not handle request if insecure request comes.
    """
    config = get_config(registry)

    ssl_redirect = config.ssl_redirect

    ignore_paths = config.ignore_paths
    if ssl_redirect.ignore_paths:
        ignore_paths = ssl_redirect.ignore_paths

    proto_header = config.proto_header
    if ssl_redirect.proto_header:
        proto_header = ssl_redirect.proto_header

    tween_name = 'ssl_redirect'

    def _ssl_redirect_tween(req):
        if not ssl_redirect.enabled:
            return handler(req)

        # ignore
        if apply_path_filter(req, ignore_paths):
            logger.info('(%s) Ignore path %s', tween_name, req.path)
            return handler(req)

        criteria = build_criteria(req, proto_header=proto_header)
        secure = all(criteria)

        if secure:
            return handler(req)

        logger.warning('(%s) Insecure request %s', tween_name, req.url)

        raise HTTPMovedPermanently(location='https://{:s}{:s}'.format(
            req.host, req.path))

    return _ssl_redirect_tween



from pyramid_secure_response.util import (
    logger,
    apply_path_filter,
    get_config,
    build_criteria,
)

HEADER_KEY = 'Strict-Transport-Security'


def build_hsts_header(config):
    """Returns HSTS Header value."""
    value = 'max-age={0}'.format(config.max_age)
    if config.include_subdomains:
        value += '; includeSubDomains'
    if config.preload:
        value += '; preload'
    return value


def tween(handler, registry):
    r"""Sets HTTP Strict Transport Security Header as configured.

    About details of HSTS, see below:

    * https://hstspreload.org/
    * https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/\
        Strict-Transport-Security#Preloading_Strict_Transport_Security
    """
    config = get_config(registry)

    hsts_support = config.hsts_support

    ignore_paths = config.ignore_paths
    if hsts_support.ignore_paths:
        ignore_paths = hsts_support.ignore_paths

    proto_header = config.proto_header
    if hsts_support.proto_header:
        proto_header = hsts_support.proto_header

    tween_name = 'hsts_support'

    def _hsts_support_tween(req):
        if not hsts_support.enabled:
            return handler(req)

        # ignore
        if apply_path_filter(req, ignore_paths):
            logger.info('(%s) Ignore path %s', tween_name, req.path)
            return handler(req)

        criteria = build_criteria(req, proto_header=proto_header)
        secure = all(criteria)

        if not secure:
            # sets the header only for https
            logger.warning('(%s) Insecure request %s', tween_name, req.url)
            return handler(req)

        res = handler(req)

        if HEADER_KEY not in res.headers:
            # ignore if already exists
            res.headers[HEADER_KEY] = build_hsts_header(hsts_support)

        return res

    return _hsts_support_tween



import pytest

from pyramid_secure_response.hsts_support import tween


@pytest.fixture(autouse=True)
def setup():
    import logging
    from pyramid_secure_response.hsts_support import logger
    logger.setLevel(logging.ERROR)


@pytest.mark.parametrize('max_age,include_subdomains,preload,header', [
    ('3600', True, True, 'max-age=3600; includeSubDomains; preload'),
    ('1800', True, False, 'max-age=1800; includeSubDomains'),
    ('900', False, False, 'max-age=900'),
])
def test_build_hsts_header(max_age, include_subdomains, preload, header):
    from collections import namedtuple
    from pyramid_secure_response.hsts_support import build_hsts_header

    Config = namedtuple('hsts_support', (  # pylint: disable=invalid-name
        'max_age',
        'include_subdomains',
        'preload',
    ))
    hsts_support = Config(max_age, include_subdomains, preload)

    assert header == build_hsts_header(hsts_support)


def test_hsts_support_tween_with_disabled(mocker, dummy_request):
    mocker.patch('pyramid_secure_response.hsts_support.apply_path_filter',
                 return_value=True)
    mocker.patch('pyramid_secure_response.hsts_support.build_criteria',
                 return_value=[])

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'False'
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count
    assert 0 == apply_path_filter.call_count
    assert 0 == build_criteria.call_count
    assert 'Strict-Transport-Security' not in res.headers


def test_hsts_support_tween_with_ignored_path(mocker, dummy_request):
    mocker.patch('pyramid_secure_response.hsts_support.apply_path_filter',
                 return_value=True)
    mocker.patch('pyramid_secure_response.hsts_support.build_criteria',
                 return_value=[])

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    dummy_request.path = '/humans.txt'
    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'True',
        'pyramid_secure_response.hsts_support.ignore_paths': '\n/humans.txt\n'
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count
    assert 1 == apply_path_filter.call_count
    apply_path_filter.assert_called_once_with(
        dummy_request, ('/humans.txt',))
    assert 0 == build_criteria.call_count
    assert 'Strict-Transport-Security' not in res.headers


def test_hsts_tween_with_none_ssl_request(mocker, dummy_request):
    from pyramid_secure_response import hsts_support
    mocker.spy(hsts_support, 'apply_path_filter')
    mocker.spy(hsts_support, 'build_criteria')

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    from pyramid_secure_response.util import get_config

    dummy_request.url = 'http://example.org/'
    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'True',
        'pyramid_secure_response.hsts_support.max_age': '31536000',
        'pyramid_secure_response.hsts_support.include_subdomains': 'True',
        'pyramid_secure_response.hsts_support.preload': 'True',
        'pyramid_secure_response.hsts_support.proto_header': '',
        'pyramid_secure_response.hsts_support.ignore_paths': '\n',
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count

    assert 1 == apply_path_filter.call_count
    apply_path_filter.assert_called_once_with(dummy_request, tuple())

    assert 1 == build_criteria.call_count
    config = get_config(dummy_request.registry)
    build_criteria.assert_called_once_with(
        dummy_request, proto_header=config.hsts_support.proto_header)

    assert 'Strict-Transport-Security' not in res.headers


def test_hsts_tween_with_ssl_request_plus_none_ssl_extra_header(
        mocker, dummy_request):
    from pyramid_secure_response import hsts_support
    mocker.spy(hsts_support, 'apply_path_filter')
    mocker.spy(hsts_support, 'build_criteria')

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    from pyramid_secure_response.util import get_config

    dummy_request.url = 'https://example.org/'
    dummy_request.headers['X-Forwarded-Proto'] = 'http'
    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'True',
        'pyramid_secure_response.hsts_support.max_age': '3600',
        'pyramid_secure_response.hsts_support.include_subdomains': 'True',
        'pyramid_secure_response.hsts_support.preload': 'True',
        'pyramid_secure_response.hsts_support.proto_header':
            'X-Forwarded-Proto',
        'pyramid_secure_response.hsts_support.ignore_paths': '\n',
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count

    assert 1 == apply_path_filter.call_count
    apply_path_filter.assert_called_once_with(dummy_request, tuple())

    assert 1 == build_criteria.call_count
    config = get_config(dummy_request.registry)
    build_criteria.assert_called_once_with(
        dummy_request, proto_header=config.hsts_support.proto_header)

    assert 'Strict-Transport-Security' not in res.headers


def test_hsts_tween_with_ssl_request(mocker, dummy_request):
    from pyramid_secure_response import hsts_support
    mocker.spy(hsts_support, 'apply_path_filter')
    mocker.spy(hsts_support, 'build_criteria')

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    from pyramid_secure_response.util import get_config

    dummy_request.url = 'https://example.org/'
    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'True',
        'pyramid_secure_response.hsts_support.max_age': '300',  # 5 minutes.
        'pyramid_secure_response.hsts_support.include_subdomains': 'True',
        'pyramid_secure_response.hsts_support.preload': 'True',
        'pyramid_secure_response.hsts_support.proto_header': '',
        'pyramid_secure_response.hsts_support.ignore_paths': '\n',
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count

    assert 1 == apply_path_filter.call_count
    apply_path_filter.assert_called_once_with(dummy_request, tuple())

    assert 1 == build_criteria.call_count
    config = get_config(dummy_request.registry)
    build_criteria.assert_called_once_with(
        dummy_request, proto_header=config.hsts_support.proto_header)

    assert 'Strict-Transport-Security' in res.headers
    assert 'max-age=300; includeSubDomains; preload' == \
        res.headers['Strict-Transport-Security']


def test_hsts_tween_with_ssl_request_plus_extra_header_check(
        mocker, dummy_request):
    from pyramid_secure_response import hsts_support
    mocker.spy(hsts_support, 'apply_path_filter')
    mocker.spy(hsts_support, 'build_criteria')

    from pyramid.response import Response
    from pyramid_secure_response.hsts_support import (
        apply_path_filter,
        build_criteria,
    )

    from pyramid_secure_response.util import get_config

    dummy_request.url = 'https://example.org/'
    dummy_request.headers['X-Forwarded-Proto'] = 'https'
    dummy_request.registry.settings = {
        'pyramid_secure_response.hsts_support.enabled': 'True',
        'pyramid_secure_response.hsts_support.max_age': '604800',  # 1 week
        'pyramid_secure_response.hsts_support.include_subdomains': 'True',
        'pyramid_secure_response.hsts_support.preload': 'True',
        'pyramid_secure_response.hsts_support.proto_header':
            'X-Forwarded-Proto',
        'pyramid_secure_response.hsts_support.ignore_paths': '\n',
    }

    handler_stub = mocker.stub(name='handler_stub')
    handler_stub.return_value = Response(status=200)
    hsts_support_tween = tween(handler_stub, dummy_request.registry)
    res = hsts_support_tween(dummy_request)

    # pylint: disable=no-member
    assert 1 == handler_stub.call_count

    assert 1 == apply_path_filter.call_count
    apply_path_filter.assert_called_once_with(dummy_request, tuple())

    assert 1 == build_criteria.call_count
    config = get_config(dummy_request.registry)
    build_criteria.assert_called_once_with(
        dummy_request, proto_header=config.hsts_support.proto_header)

    assert 'Strict-Transport-Security' in res.headers
    assert 'max-age=604800; includeSubDomains; preload' == \
        res.headers['Strict-Transport-Security']



from collections import namedtuple
import logging

from pyramid_secure_response import __name__ as PACKAGE_NAME


logger = logging.getLogger(PACKAGE_NAME)  # pylint: disable=invalid-name


def _get_config_value_f(registry, config_key=''):
    # type: (Registry, str) -> 'function'
    """Gets configured value from .ini file via registry."""
    if not config_key:
        raise ValueError

    s = registry.settings

    def _get_value_f(key, default):
        v = s.get('{:s}.{:s}'.format(config_key, key), default)
        if v == default:
            return v
        v = str(v)
        if v.lower() == 'true':
            return True
        elif v.lower() == 'false':
            return False
        if '\n' in v:
            return tuple(v.split())

        return v

    return _get_value_f


def _build_config(prefix='', defaults=tuple(), registry=None):
    # pylint: disable=invalid-name
    if prefix:
        config_key = '{:s}.{:s}'.format(PACKAGE_NAME, prefix)
    else:
        config_key = PACKAGE_NAME

    get_value_f = _get_config_value_f(registry, config_key=config_key)

    Config = namedtuple('Config', [k for k, _ in defaults])
    return Config(**dict([(k, get_value_f(k, v)) for k, v in defaults]))


def get_config(registry):  # type: (Registry) -> namedtuple
    """Returns namedtuple instance object has configuration."""
    # HTTP Redirections (ssl_redirect.xxx)
    ssl_redirect = _build_config(prefix='ssl_redirect', defaults=(
        ('enabled', True),
        ('proto_header', ''),
        ('ignore_paths', tuple()),
    ), registry=registry)

    # HTTP Strict Transport Security (hsts_support.xxx)
    hsts_support = _build_config(prefix='hsts_support', defaults=(
        ('enabled', True),
        ('proto_header', ''),
        ('ignore_paths', tuple()),
        ('max_age', '31536000'),  # seconds, 1 year
        ('include_subdomains', True),
        ('preload', True),
    ), registry=registry)

    # Shared
    return _build_config(prefix='', defaults=(
        ('proto_header', ''),   # e.g. X-Forwarded-Proto
        ('ignore_paths', tuple()),

        ('ssl_redirect', ssl_redirect),
        ('hsts_support', hsts_support),
    ), registry=registry)


def apply_path_filter(req, paths):  # type: (Request, tuple) -> bool
    if paths:
        return any([req.path.startswith(path) for path in paths])
    return False


def build_criteria(req, **kwargs):  # type: (Request, dict) -> tuple
    """Builds criteria contains about incoming request.

    This function is used for following tweens:

    * ssl_redirect
    * hsts_support
    """
    criteria = [
        req.url.startswith('https://'),
    ]
    if 'proto_header' in kwargs:
        proto_header = kwargs.get('proto_header', None)
        if proto_header:
            criteria.append(req.headers.get(proto_header, 'http') == 'https')

    return tuple(criteria)



import pytest


@pytest.fixture(scope='function')
def dummy_request():  # type: () -> Request
    from pyramid import testing

    url = 'http://example.org'
    settings = {}
    req = testing.DummyRequest(
        environ={},
        locale_name='en',
        matched_route=None,
        settings=settings,
        server_name='example.org',
        subdomain='',
        host='example.org:80',
        application_url=url,
        url=url,
        host_url=url,
        path_url=url,
    )
    req.registry.settings = settings
    return req
